/*
 * Copyright 2005 Luther E. Birdzell.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

package com.itko.lisa.ext.util;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.OutputStreamWriter;
import java.io.Reader;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.apache.commons.io.IOUtils;
import org.apache.commons.lang.StringUtils;

import com.itko.lisa.ext.exception.MissingParameterException;

/**
 * Utility methods for working with the file system through the java.io API.
 * 
 * @author <a href="mailto:luther@dctmutils.org">Luther E. Birdzell</a>
 */
public class FileHelper
{
    private static Logger log = Logger.getLogger(FileHelper.class.getName());

    /**
     * All methods are static.
     */
    protected FileHelper()
    {
    }

    /**
     * Recursively remove all content and dirs under the specified path.
     * 
     * @param path
     */
    public static void recursiveRemoveDirectory(File path)
    {
        if ( path == null )
        {
            throw new MissingParameterException("path");
        }

        log.info("recursiveRemoveDirectory called with "
                + path.getAbsolutePath());

        recursiveEmptyDirectory(path);

        File[] contents = path.listFiles();
        File dirToDelete = null;

        if ( contents != null )
        {
            for ( int i = 0; i < contents.length; i++ )
            {
                dirToDelete = contents[i];
                log.fine("deleting dir " + dirToDelete.getName());

                if ( !dirToDelete.delete() )
                {
                    log.warning("Failed to delete: "
                            + dirToDelete.getAbsolutePath());
                }
                recursiveRemoveDirectory(contents[i]);
            }
        }

        path.delete();
    }

    /**
     * Recursively remove the files in the specified path.
     * 
     * @param path
     */
    public static void recursiveEmptyDirectory(File path)
    {
        if ( path == null )
        {
            throw new MissingParameterException("path");
        }

        log.fine("recursiveEmptyDirectory called with "
                + path.getAbsolutePath());

        File[] contents = path.listFiles();
        File fileToDelete = null;

        if ( contents != null )
        {
            for ( int i = 0; i < contents.length; i++ )
            {
                if ( contents[i].isFile() )
                {
                    fileToDelete = contents[i];
                    log.fine("deleting file " + fileToDelete.getName());

                    fileToDelete.delete();
                }
            }
        }
    }

    /**
     * Returns true if the last three characters of a file name are 'xml'
     * 
     * @param fileName
     * @return a <code>boolean</code> value
     */
    public static boolean isXmlFile(String fileName)
    {
        if ( StringUtils.isBlank(fileName) )
        {
            throw new MissingParameterException("fileName");
        }

        boolean isXmlFile = false;

        String fileExtension = fileName.substring(fileName.length() - 3,
                                                  fileName.length());

        isXmlFile = (fileExtension.equalsIgnoreCase("xml") ? true : false);

        return isXmlFile;
    }

    /**
     * Replace back slashes '\' with slashes '/'.
     * 
     * @param path
     * @return a <code>String</code> value
     */
    public static String replaceBackSlashWithSlash(String path)
    {
        if ( StringUtils.isBlank(path) )
        {
            throw new MissingParameterException("path");
        }

        path = StringUtils.replace(path, "\\", "/");

        return path;
    }

    /**
     * Remove the fileName from the path.
     * 
     * @param path
     * @return a <code>String</code> value
     */
    public static String removeFileName(String path)
    {
        if ( StringUtils.isBlank(path) )
        {
            throw new MissingParameterException("path");
        }

        path = replaceBackSlashWithSlash(path);
        path = path.substring(0, path.lastIndexOf("/"));

        return path;
    }

    /**
     * Get the fileName from the path.
     * 
     * @param path
     * @return a <code>String</code> value
     */
    public static String getFileName(String path)
    {
        if ( StringUtils.isBlank(path) )
        {
            throw new MissingParameterException("path");
        }

        String fileName = null;

        path = replaceBackSlashWithSlash(path);

        // remove the path from the file
        if ( path.indexOf("/") != -1 )
        {
            fileName = path.substring(path.lastIndexOf("/") + 1, path.length());
        }
        else
        // only a file name was passed in
        {
            fileName = path;
        }

        log.fine("fileName = " + fileName);

        return fileName;
    }

    /**
     * Remove the file extension from the file name.
     * 
     * @param fileName
     * @return a <code>String</code> value
     */
    public static String removeFileExtension(String fileName)
    {
        if ( StringUtils.isBlank(fileName) )
        {
            throw new MissingParameterException("fileName");
        }

        if ( fileName.indexOf(".") > 0 )
        {
            fileName = fileName.substring(0, fileName.lastIndexOf("."));
        }

        return fileName;
    }

    /**
     * Get the file extension from the file name. Returns empty
     * <code>String</code> if the fileName does not have an extension.
     * 
     * @param fileName
     * @return a <code>String</code> value
     */
    public static String getFileExtension(String fileName)
    {
        if ( StringUtils.isBlank(fileName) )
        {
            throw new MissingParameterException("fileName");
        }

        String extension = null;

        if ( fileName.indexOf(".") > 0 )
        {
            extension = fileName.substring(fileName.lastIndexOf(".") + 1,
                                           fileName.length());
        }

        return extension;
    }

    /**
     * Converts a text <code>File</code> to a String.
     * 
     * @param file
     * @return a <code>String</code> value
     * @exception IOException
     */
    public static String getFileAsString(File file) throws IOException
    {
        if ( file == null )
        {
            throw new MissingParameterException("file");
        }

        FileReader reader = new FileReader(file);

        String fileContents = getFileAsString(reader);

        reader.close();

        return fileContents;
    }

    /**
     * Convert a text file to a <code>String</code>.
     * 
     * @param path
     * @return a <code>String</code> value
     * @exception IOException
     */
    public static String getFileAsString(String path) throws IOException
    {
        if ( StringUtils.isBlank(path) )
        {
            throw new MissingParameterException("path");
        }

        FileReader reader = new FileReader(path);

        String fileContents = getFileAsString(reader);

        reader.close();

        return fileContents;
    }

    /**
     * Use the default ClassLoader to load the file, then convert it to a
     * <code>String</code>.
     * 
     * @param fileName
     * @return a <code>String</code> value
     * @exception IOException
     */
    public static String getFileAsStringFromClassPath(String fileName)
        throws IOException
    {
        if ( StringUtils.isBlank(fileName) )
        {
            throw new MissingParameterException("fileName");
        }

        InputStream is = getFileAsStreamFromClassPath(fileName);

        InputStreamReader reader = null;

        if ( is != null )
        {
            reader = new InputStreamReader(is);
        }
        else
        {
            throw new RuntimeException(fileName
                    + " was not found in the CLASSPATH.");
        }

        String fileContents = getFileAsString(reader);

        reader.close();

        return fileContents;
    }

    /**
     * Use the default ClassLoader to load the file.
     * 
     * @param fileName
     * @return a <code>String</code> value
     * @exception IOException
     */
    public static InputStream getFileAsStreamFromClassPath(String fileName)
        throws IOException
    {
        if ( StringUtils.isBlank(fileName) )
        {
            throw new MissingParameterException("fileName");
        }

        InputStream instream = FileHelper.class.getResourceAsStream((fileName.startsWith("/") ? ""
                : "/")
                + fileName);

        return instream;
    }

    /**
     * Use the default ClassLoader to load the file.
     * 
     * @param fileName
     * @return a <code>String</code> value
     * @exception IOException
     */
    public static byte[] getFileAsBytesFromClassPath(String fileName)
        throws IOException
    {
        InputStream instream = getFileAsStreamFromClassPath(fileName);

        byte[] bytes = IOUtils.toByteArray(instream);

        instream.close();

        return bytes;
    }

    /**
     * Get the file as a byte[]
     * 
     * @param fileName
     * @return a <code>String</code> value
     * @exception IOException
     */
    public static byte[] getFileAsBytes(String fileName) throws IOException
    {
        InputStream instream = new FileInputStream(new File(fileName));

        byte[] bytes = IOUtils.toByteArray(instream);

        instream.close();

        return bytes;
    }

    /**
     * Convert the file to a string.
     * 
     * @param reader
     * @return a <code>String</code> value
     * @exception IOException
     */
    protected static String getFileAsString(Reader reader) throws IOException
    {
        BufferedReader br = new BufferedReader(reader);
        StringBuffer buff = new StringBuffer();
        String line = null;

        do
        {
            line = br.readLine();

            if ( line != null )
            {
                buff.append(line);
            }

        }
        while ( line != null );

        String testFile = buff.toString();
        // logger.debug("testFile: " + testFile);

        return testFile;
    }

    /**
     * Convert the input <code>String</code> to camel case with the first letter
     * upper case. E.x. input 'foo Bar' --> 'FooBar'
     * 
     * @param words
     * @return a camel case <code>String</code>
     */
    public static String camelCaseWithFirstLetterUpperCase(String words)
    {
        // uppercase is the default behavior of the protected method
        String camelCase = camelCase(words);
        return camelCase;
    }

    /**
     * Convert the input <code>String</code> to camel case with the first letter
     * lower case. E.x. input 'foo Bar' --> 'fooBar'
     * 
     * @param words
     * @return a camel case <code>String</code>
     */
    public static String camelCaseWithFirstLetterLowerCase(String words)
    {
        String camelCase = camelCase(words);

        if ( StringUtils.isNotBlank(camelCase) )
        {
            String lowerCaseFirstLetter = StringUtils.substring(StringUtils.lowerCase(camelCase),
                                                                0,
                                                                1);

            StringBuffer buff = new StringBuffer();
            buff.append(lowerCaseFirstLetter)
                .append(StringUtils.substring(camelCase, 1, words.length()));

            camelCase = buff.toString();
        }

        return camelCase;
    }

    /**
     * Convert the input <code>String</code> to camel case.
     * 
     * @param words
     * @return a camel case <code>String</code>
     */
    protected static String camelCase(String words)
    {
        log.fine("words = " + words);

        if ( StringUtils.isBlank(words) )
        {
            return words;
        }

        int pos = 0;

        words = StringUtils.replace(words, "-", " ");
        words = StringUtils.replace(words, "_", " ");

        words = words.substring(words.lastIndexOf('.') + 1);
        words = words.substring(words.lastIndexOf('$') + 1);
        char[] chars = words.toCharArray();

        StringBuffer buf = new StringBuffer();

        while ( pos < chars.length )
        {
            if ( Character.isWhitespace(chars[pos]) )
            {
                pos++;
                if ( pos < chars.length )
                {
                    buf.append(Character.toUpperCase(chars[pos]));
                }
            }
            else if ( pos == 0 )
            {
                buf.append(Character.toUpperCase(chars[pos]));
            }
            else
            {
                buf.append(chars[pos]);
            }
            pos++;
        }

        log.fine("words = " + buf.toString());
        return buf.toString();
    }

    /**
     * Write the specified String to the specified path with a UTF-8 character
     * encoding.
     * 
     * @param body
     * @param path
     */
    public static void writeStringToFileSystem(String body, String path)
    {
        FileOutputStream fos = null;
        OutputStreamWriter osw = null;

        try
        {
            fos = new FileOutputStream(path);
            osw = new OutputStreamWriter(fos, "UTF-8");
            osw.write(body);
            osw.flush();
        }
        catch ( Exception e )
        {
            log.log(Level.SEVERE, e.getMessage(), e);
            throw new RuntimeException(e.getMessage(), e);
        }
        finally
        {
            IOUtils.closeQuietly(fos);
            IOUtils.closeQuietly(osw);
        }
    }

    /**
     * 
     * 
     * @param path
     * @return the output path with a trailing <code>File.separator</code>
     */
    public static String ensureTrailingFileSeparator(String path)
    {
        if ( StringUtils.isNotBlank(path)
                && StringUtils.lastIndexOf(path, File.separator) != (path.length() - 1) )
        {
            path += File.separator;
        }
        return path;
    }

    /**
     * 
     * 
     * @param url
     * @return the output path with a trailing slash '/'
     */
    public static String ensureTrailingSlash(String url)
    {
        if ( StringUtils.isNotBlank(url)
                && StringUtils.lastIndexOf(url, "/") != (url.length() - 1) )
        {
            url += "/";
        }
        return url;
    }

    /**
     * @param obj
     * @param file
     * @throws IOException
     */
    public static void saveToSer(Object obj, File file) throws IOException
    {

        if ( file.exists() && !file.delete() )
        {
            throw new IOException("Unsuccessful deleting file: "
                    + file.getAbsolutePath());
        }
        file.createNewFile();
        ObjectOutputStream oos = new ObjectOutputStream(new FileOutputStream(file));
        oos.writeObject(obj);
        oos.close();
    }

    /**
     * @param file
     * @return
     * @throws ClassNotFoundException
     * @throws IOException
     */
    public static Object getFromSer(File file)
        throws IOException, ClassNotFoundException
    {
        ObjectInputStream ois = new ObjectInputStream(new FileInputStream(file));
        Object o = ois.readObject();
        ois.close();
        return o;
    }

    /**
     * 
     * 
     * @param in
     * @return
     */
    public static String stripNonAscii(String in)
    {
        if ( in == null )
        {
            return null;
        }

        StringBuilder clean = new StringBuilder();

        for ( int i = 0; i < in.length(); i++ )
        {

            char c = in.charAt(i);

            int cn = ( int ) c;

            /*
             * characters from 32 (space) to 127 (tilde) are old traditional
             * 7-bit-clean ASCII and will work. Unicode characters from 128
             * upwards should all work if converted to numeric entities apart
             * from the block 0xd800-0xdfff (surrogates). Valid Unicode
             * character codes end at 0x10ffff
             */
            if ( ((cn >= 32) && (cn < 127)) || (cn == 9) || (cn == 10)
                    || (cn == 9) )
            {
                clean.append(c);
            }

        }

        return clean.toString();

    }
}
