package com.itko.lisa.ext.util;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.util.Date;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.apache.commons.io.IOUtils;

/**
 * Utility class used to manage a time-based filesystem mutex. This utility can
 * be used to prevent a test from running more than once in a specified time
 * interval.
 * 
 * @author <a href="mailto:luther@itko.com">Luther Birdzell</a>
 */
public class FileSystemMutexHelper
{
    private static Logger log = Logger.getLogger(FileSystemMutexHelper.class.getName());

    /**
     * 
     * 
     * @param testCaseName
     * @param testExecutionIntervalSeconds
     *            - the minimum number of seconds to wait between test
     *            executions
     * @return a boolean indicating whether the test can run or not
     */
    public static boolean canTestRun(String testCaseName,
                                     int testExecutionIntervalSeconds)
    {
        boolean canTestRun = false;

        try
        {
            File lockFile = new File("/tmp" + testCaseName);

            if ( lockFile.exists() )
            {
                FileInputStream input = new FileInputStream(lockFile);
                List lines = IOUtils.readLines(input);
                IOUtils.closeQuietly(input);

                String timeOfLastTestString = ( String ) lines.get(0);
                long timeOfLastTest = Long.parseLong(timeOfLastTestString);
                long currentTime = System.currentTimeMillis();
                log.fine("timeOfLastTest = " + new Date(timeOfLastTest)
                        + ", currentTime = " + new Date(currentTime));

                // if the test ran less than 1 minute ago
                if ( currentTime - timeOfLastTest > (testExecutionIntervalSeconds * 1000) )
                {
                    log.info("the test ran over "
                            + testExecutionIntervalSeconds
                            + "  seconds ago, running test");

                    canTestRun = true;

                    FileOutputStream output = new FileOutputStream(lockFile);
                    IOUtils.write(String.valueOf(currentTime), output);
                    IOUtils.closeQuietly(output);
                }
                else
                {
                    log.info("the test already ran within the "
                            + testExecutionIntervalSeconds
                            + " second window, not running...");
                }
            }
            else
            {
                log.info("lockfile did not exist running test");

                canTestRun = true;

                FileOutputStream output = new FileOutputStream(lockFile);
                IOUtils.write(String.valueOf(System.currentTimeMillis()),
                              output);
                IOUtils.closeQuietly(output);
            }
        }
        catch ( Exception e )
        {
            log.log(Level.SEVERE, e.getMessage(), e);
            throw new RuntimeException(e.getMessage(), e);
        }

        return canTestRun;
    }
}
