/*
 * Copyright 2005 Luther E. Birdzell.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

package com.itko.lisa.ext.util;

import java.util.Enumeration;
import java.util.HashMap;
import java.util.Locale;
import java.util.Map;
import java.util.MissingResourceException;
import java.util.ResourceBundle;
import java.util.logging.Logger;

import org.apache.commons.lang.StringUtils;

import com.itko.lisa.ext.exception.MissingParameterException;

/**
 * This class facilitates reading properties from a properties file. Pass the
 * property file name without the '.properties' extension to the constructor.
 * Note: the directory in which the properties file is stored, must be in the
 * application's classpath.
 * 
 * @author <a href="mailto:luther@dctmutils.org">Luther E. Birdzell</a>
 * @author <a href="mailto:dbingham@users.sourceforge.net">Daniel Bingham</a>
 */
public class PropertiesReader
{
    private static Logger  log    = Logger.getLogger(PropertiesReader.class.getName());

    private ResourceBundle bundle = null;

    /**
     * Creates a new <code>PropertiesReader</code> instance.
     * 
     * @param propertyFileName
     */
    public PropertiesReader(String propertyFileName)
    {
        log.fine("propertyFileName = " + propertyFileName);
        this.bundle = ResourceBundle.getBundle(propertyFileName);
    }

    /**
     * Creates a new <code>PropertiesReader</code> instance for a specific
     * <code>Locale</code>.
     * 
     * @param propertyFileName
     * @param locale
     */
    public PropertiesReader(String propertyFileName, Locale locale)
    {
        log.fine("propertyFileName = " + propertyFileName);
        this.bundle = ResourceBundle.getBundle(propertyFileName, locale);
    }

    /**
     * Returns the value associated with the key in the properties file that was
     * passed to the constructor.
     * 
     * @param key
     * @return a <code>String</code> value
     */
    public String getValue(String key)
    {
        String value = null;

        if ( StringUtils.isNotBlank(key) )
        {
            try
            {
                value = bundle.getString(key);
            }
            catch ( MissingResourceException ignore )
            {
                log.warning("key " + key + " not found in ResourceBundle");
            }
        }

        return value;
    }

    /**
     * Returns the <code>boolean</code> value associated with the key in the
     * properties file that was passed to the constructor.
     * 
     * @param key
     * @return a <code>boolean</code> value false, unless the key is set to
     *         'true'
     */
    public boolean getBooleanValue(String key)
    {
        boolean value = false;

        String valueStr = getValue(key);

        if ( StringUtils.isNotBlank(valueStr) )
        {
            value = new Boolean(valueStr).booleanValue();
        }

        return value;
    }

    /**
     * Returns the <code>int</code> value associated with the key in the
     * properties file that was passed to the constructor.
     * 
     * @param key
     * @param defaultValue
     * @return an <code>int</code> value or the <code>defaultValue</code> if
     *         there was a problem reading the key.
     */
    public int getIntValue(String key, int defaultValue)
    {
        int value = defaultValue;

        String valueStr = getValue(key);

        if ( StringUtils.isNotBlank(valueStr) )
        {
            try
            {
                value = Integer.parseInt(valueStr);
            }
            catch ( NumberFormatException nfe )
            {
                log.warning("Error parsing the value '" + valueStr
                        + "' from the properties file");
            }
        }

        return value;
    }

    /**
     * Returns the value associated with the key in the properties file that was
     * passed to the constructor.
     * 
     * Throws a <code>MissingParameterException</code> if the value is not set.
     * 
     * @param key
     * @return a <code>String</code> value
     */
    public String getRequiredValue(String key)
    {
        String value = getValue(key);

        if ( StringUtils.isBlank(value) )
        {
            throw new MissingParameterException(key);
        }

        return value;
    }

    /**
     * Get all of the values from the properties file as a map.
     * 
     * @return a <code>Map</code> object containing all of the values from the
     *         proprties file.
     */
    public Map getValuesAsMap()
    {
        Map props = new HashMap();

        Enumeration keys = bundle.getKeys();
        while ( keys.hasMoreElements() )
        {
            String key = ( String ) keys.nextElement();
            props.put(key, getValue(key));
        }

        return props;
    }
}
