package com.itko.lisa.ext.util;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Logger;

import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

import com.itko.lisa.editor.TestCaseTypeData;
import com.itko.lisa.model.mar.ModelArchive;
import com.itko.lisa.model.mar.ModelArchiveInfo;
import com.itko.lisa.repo.ProjectManager;
import com.itko.lisa.test.CompanionInterface;
import com.itko.lisa.test.Environment;
import com.itko.lisa.test.FilterInterface;
import com.itko.lisa.test.Project;
import com.itko.lisa.test.TestCase;
import com.itko.lisa.test.TestCaseDocument;
import com.itko.lisa.test.TestDefException;
import com.itko.lisa.test.TestExec;
import com.itko.lisa.test.TestNode;
import com.itko.lisa.vse.stateful.ConversationalStep;
import com.itko.util.ParameterList;

/**
 * TestCaseHelper - Helper class for working with test cases
 *
 * @author <a href="mailto:daniel.bingham@itko.com">Daniel Bingham</a>
 */
public class TestCaseHelper
{
    private static final Logger log = Logger.getLogger( TestCaseHelper.class.getName() );

    public static final String VSM_IMAGE_RESPONSE_SELECTION_STEP_NAME = "VS Image Response Selection";

    /**
     * Method getCompanionNodes gets the companions on the test case that the calling step is on.
     *
     * @param stepNode of type Node is the node that represents the step calling this function.
     *
     * @return List<Node> - The list of companions if any.  If no companions are present, it will be an empty list.  If
     *         null is passed for stepNode, null will be returned.
     */
    public static List<Node> getCompanionNodes( Node stepNode )
    {
        if( stepNode == null )
        {
            return null;
        }

        NodeList nodes = stepNode.getParentNode().getChildNodes();
        List<Node> companions = new ArrayList<Node>();
        for( int i = 0; i < nodes.getLength(); i++ )
        {
            Node tmpNode = nodes.item( i );
            if( "Companion".equals( tmpNode.getNodeName() ) )
            {
                companions.add( tmpNode );
            }
        }

        return companions;
    }

    /**
     * Method getCompanion gets the companion of the specified type from the test case that the calling step is on.
     *
     * @param typeName of type String is the fully qualified class name of the companion type to search for
     * @param stepNode of type Node is the node that represents the step calling this function.
     *
     * @return Node - The node representing the companion on the test case.  If none are found that match, null is
     *         returned.  If null is provided for either argument, null is returned.
     */
    public static Node getCompanionNode( String typeName, Node stepNode )
    {
        if( typeName == null || stepNode == null )
        {
            return null;
        }

        List<Node> companions = getCompanionNodes( stepNode );
        for( Node companion : companions )
        {
            if( typeName.equals( companion.getAttributes().getNamedItem( "type" ).getNodeValue() ) )
            {
                return companion;
            }
        }

        return null;
    }

    /**
     * Method getCompanionInterfaces gets the List of CompanionInterface objects off of the TestCase object
     *
     * @param tc of type TestCase is the TestCase to search
     *
     * @return List<CompanionInterface> the List of CompanionInterface objects from the Test case.  Empty List is
     *         returned if none found.  If null is provided for the TestCase, null is returned.
     */
    public static List<CompanionInterface> getCompanionInterfaces( TestCase tc )
    {
        if( tc == null )
        {
            return null;
        }

        return tc.getCompanions();
    }

    /**
     * Method getCompanionInterface gets the CompanionInterface object for the companion of the specified type on the
     * TestCase
     *
     * @param tc       of type TestCase is the TestCase to search
     * @param typeName of type String is the fully qualified class name of the companion type to search for
     *
     * @return CompanionInterface is the CompanionInterface object for the companion requested.  null if not found.
     *         null
     *         if null is provided for either argument.
     */
    public static CompanionInterface getCompanionInterface( TestCase tc, String typeName )
    {
        //early out
        if( tc == null || typeName == null )
        {
            return null;
        }

        List<CompanionInterface> companions = getCompanionInterfaces( tc );
        for( CompanionInterface companion : companions )
        {
            if( companion.getClass().getName().equals( typeName ) )
            {
                return companion;
            }
        }

        return null;
    }

    public static Map<String, ParameterList> getRequestDataHandlers( TestCase tc )
    {
        Map<String, Map<String, ParameterList>> allDPhs = getDataHandlers( tc );

        return allDPhs.get( "request" );
    }

    public static Map<String, ParameterList> getResponseDataHandlers( TestCase tc )
    {
        Map<String, Map<String, ParameterList>> allDPHs = getDataHandlers( tc );

        return allDPHs.get( "response" );
    }

    /**
     * Gets all the data handlers for the VSM at the specified path.
     *
     * The returned Map contains two entries, one with the key "request", one with the key "response".  The two entries
     * correspond to the request and response side data protocols respectively.  Each entry is, itself, a Map where the
     * keys are the DataProtocol class names and the values are the parameters for the data protocol, if any.  The Map
     * implementation used DOES preserve order, so the execution order of the data protocols will be maintained.
     *
     * @param pathToVSM is the path to the VSM to parse
     *
     * @return a Map<String,Map<String, ParameterList>> as described above.
     * @throws IOException
     * @throws SAXException
     * @throws TestDefException
     */
    public static Map<String, Map<String, ParameterList>> getDataHandlers( String pathToVSM )
    throws IOException, SAXException, TestDefException
    {
        TestCaseDocument tcDoc = new TestCaseDocument( pathToVSM, TestCaseTypeData.VSM_FILE_EXTENSION );

        TestCase tc = TestCase.loadTestCase( tcDoc, null, TestCaseTypeData.Type.VS_MODEL );

        return getDataHandlers( tc );
    }

    /**
     * Gets all the data handlers for the VSM represented by the specified TestCase object.
     *
     * The returned Map contains two entries, one with the key "request", one with the key "response".  The two entries
     * correspond to the request and response side data protocols respectively.  Each entry is, itself, a Map where the
     * keys are the DataProtocol class names and the values are the parameters for the data protocol, if any.  The Map
     * implementation used DOES preserve order, so the execution order of the data protocols will be maintained.
     *
     * @param tc the TestCase object representing the VSM
     *
     * @return a Map<String,Map<String, ParameterList>> as described above.
     */
    public static Map<String, Map<String, ParameterList>> getDataHandlers( TestCase tc )
    {
        Map<String, ParameterList> requestDataHandlers =
        Collections.synchronizedMap( new LinkedHashMap<String, ParameterList>() );
        Map<String, ParameterList> responseDataHandlers =
        Collections.synchronizedMap( new LinkedHashMap<String, ParameterList>() );

        List<FilterInterface> filters = getAllFiltersFromTC( tc );
        /*
        for( FilterInterface filter : filters )
        {
            if( filter instanceof DataProtocolStandardControlFilter )
            {
                DataProtocolStandardControlFilter dphFilter = (DataProtocolStandardControlFilter) filter;
                String dph = dphFilter.getDataProtocol();
                boolean request = dphFilter.isProcessRequests();
                boolean response = dphFilter.isProcessResponses();
                ParameterList params = null;

                if( filter instanceof DataProtocolDynamicParametersControlFilter )
                {
                    String dpParams =
                    filter.getParameters().get( DataProtocolDynamicParametersControlFilter.DP_PARAMETERS );
                    params = new ParameterList( dpParams );
                }

                if( request )
                {
                    requestDataHandlers.put( dph, params );
                }

                if( response )
                {
                    responseDataHandlers.put( dph, params );
                }
            }
        }
		*/
        Map<String, Map<String, ParameterList>> dataHandlers = new HashMap<String, Map<String, ParameterList>>();
        dataHandlers.put( "request", requestDataHandlers );
        dataHandlers.put( "response", responseDataHandlers );
        throw new UnsupportedOperationException();
        //return null;
    }

    /**
     * Gets a List (in order) of all the filters found on the specified TestCase.  Note, this includes all filters
     * found
     * on all of the steps, in addition to any "global" filters on the test case itself.
     *
     * @param tc the TestCase object to inspect
     *
     * @return the List of all the filters.
     */
    public static List<FilterInterface> getAllFiltersFromTC( TestCase tc )
    {
        List<FilterInterface> filters = new ArrayList<FilterInterface>();

        filters.addAll( tc.getFiltersAsList() );

        for( String nodeName : tc.getNodes().keySet() )
        {
            filters.addAll( tc.getNodes().get( nodeName ).getFiltersAsList() );
        }

        return filters;
    }

    /**
     * Given a VSM test case, find the path to the SI referenced by it.
     *
     * @param vsm - a TestCase object representing the VSM to search
     *
     * @return the name of the SI or null if one could not be found
     */
    public static String getSIPathFromVSM( TestCase vsm )
    {
        String si = null;

        Map<String, TestNode> nodes = vsm.getNodes();
        TestNode siNode = null;
        if( nodes.containsKey( VSM_IMAGE_RESPONSE_SELECTION_STEP_NAME ) )
        {
            TestNode tmp = nodes.get( VSM_IMAGE_RESPONSE_SELECTION_STEP_NAME );
            if( tmp instanceof ConversationalStep )
            {
                siNode = tmp;
            }
        }

        if( siNode == null )
        {
            for( TestNode node : nodes.values() )
            {
                if( node instanceof ConversationalStep )
                {
                    siNode = node;
                    break;
                }
            }
        }

        if( siNode != null )
        {
            si = ((ConversationalStep) siNode).getVSISource();
            si = TestExec.parseInState( si, vsm.getLaunchState() );
        }

        return si;
    }

    public static void main( String[] args )
    {
        if( args.length == 0 )
        {
            System.out.println( "Must provide LISA_HOME an argument" );
            return;
        }
        String lisaHome = args[0];
        System.setProperty( "LISA_HOME", lisaHome );
        new Environment( "TestCaseHelper", Environment.APP_OTHER );

        try
        {
            Map<String, Map<String, ParameterList>> dphs = getDataHandlers(
                                                                          "/Users/dbingham/lisa-extensions/customer/sprint/LisaRMSProject/VServices/QueryReferenceInfoService.vsm" );
            System.out.println( "DPHs: " + dphs );
        }
        catch( Exception e )
        {
            System.err.println( "Exception: " + e.getMessage() );
            e.printStackTrace();
        }

    }

    /**
     * @param vsmFile
     *
     * @return
     * @throws com.itko.lisa.test.TestDefException
     * @throws java.io.IOException
     * @throws org.xml.sax.SAXException
     */
    public static TestCase loadVsm( File vsmFile ) throws TestDefException, IOException, SAXException
    {
        return TestCase.loadTestCase( TestCase.openTestCaseDoc( vsmFile.getAbsolutePath(), TestCaseTypeData.Type.VS_MODEL ),
                                      null,
                                      TestCaseTypeData.Type.VS_MODEL );
    }

    public static TestCase loadVsm( InputStream vsmIS ) throws IOException, SAXException, TestDefException
    {
        return TestCase.loadTestCase( TestCase.openTestCaseDoc( vsmIS, null ),
                                      null,
                                      TestCaseTypeData.Type.VS_MODEL );
    }

    public static ModelArchive buildMarForVSM( final String vsmFile,
                                               final String configFile,
                                               final int capacity,
                                               final int thinkScale,
                                               final boolean autoRestart )
    {
        ModelArchiveInfo mari = ModelArchiveInfo.createVSMARInfo( vsmFile,
                                                                  configFile,
                                                                  capacity,
                                                                  thinkScale,
                                                                  autoRestart );

        Project proj = null;
        try
        {
            proj = ProjectManager.loadProject( new File( ProjectManager.getProjectRoot( vsmFile ) ) );
        }
        catch( NullPointerException ignore ) {}

        if( mari == null || proj == null )
        {
            return null;
        }

        return new ModelArchive( proj, mari );
    }
}
