package com.itko.lisa.ext.util;

import java.io.File;
import java.rmi.RemoteException;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import com.google.common.collect.Lists;
import com.google.common.collect.MapMaker;
import com.google.common.collect.Sets;
import com.itko.lisa.coordinator.TestRegistry;
import com.itko.lisa.coordinator.VirtualService;
import com.itko.lisa.coordinator.VirtualServiceEnvironment;
import com.itko.lisa.model.mar.ModelArchive;
import com.itko.lisa.net.NetConstants;
import com.itko.lisa.test.Environment;
import com.itko.lisa.vse.VirtualServiceStatus;
import com.itko.lisa.vse.stateful.model.ServiceImage;
import com.itko.lisa.vse.stateful.protocol.DataProtocol;
import com.itko.lisa.vse.stateful.protocol.ProtocolInfo;
import com.itko.lisa.vse.stateful.protocol.TransportProtocol;
import com.itko.util.FileHelper;
import com.itko.util.ParameterList;

/**
 * Invoke a LISA test through the lisa-invoke web app.
 *
 * @author <a href="mailto:luther@itko.com">Luther E. Birdzell</a>
 * @author <a href="mailto:daniel.bingham@itko.com">Daniel Bingham</a>
 */
public final class VseHelper
{
    /** java util logger */
    private static final Logger logger = Logger.getLogger( VseHelper.class.getName() );

    /** Map of helpers for VSEs */
    private static final Map<String, VseHelper> helpers = (new MapMaker()).softValues().makeMap();

    /**
     * Calls getInstance(String) with the default VSE name, which is, ultimately, fetched by calling:
     * System#getProperty( "lisa.vseName", "VSE" );
     *
     * @return a VseHelper
     */
    public static synchronized VseHelper getInstance()
    {
        return getInstance( VirtualServiceEnvironment.DEF_VSE_NAME );
    }

    /**
     * Get an instance of the VirtualServiceImageHelper for a the specified VSE.  The name argument
     * follows the same rules as for TestRegistry#getVirtualServiceEnvironment(String).  Docs from there reproduced
     * here for reference:
     *
     * <b>Naming</b> In general, where the API has a named server in the method
     * signature, you can pass a qualified name. A  qualified name is of the
     * form serverName@labName, for example Sim1@perfLab,
     * Coordinator@funcLab, VSE@devLab. If there is no '@' in the name the Default
     * lab is implied. The Default lab is always called Default and is always
     * available.
     * <p/>
     * Names of services can be in the simple form, eg 'Coordinator' or in fully
     * qualified for, eg tcp://machine.domain:port/name. If they are in the simple
     * form, the DCM will return the first matching server name regardless of the
     * protocol, machine name or port. So, it's theoretically possible to have two
     * simulators, both called MySim but on different hosts:
     * tcp://192.168.0.1:2014/MySim and tcp://192.168.0.1:2014/MySim. They both
     * belong to a lab called PerfLab.
     * <p/>
     * If you called getSimulator("MySim@PerfLab") you would get back one of the
     * simulators, which one however is undefined. The order might be round robin,
     * or based on the simulator load profile and you get the one doing the least
     * work, just don't rely on this behavior staying constant across LISA releases.
     * <p/>
     * If you called TestRegistry#getCoordinatorServer(String) and passed null or empty
     * string, you would get a random coordinator server from the Default lab. If
     * you called getCoordinatorServer("@PerfLab") you would get a random
     * coordServer from the PerfLab lab. If you called
     * getCoordinatorServer("MyCS@PerfLab") you would get the 'MyCS' coordServer
     * from the PerfLab lab (if there *is* a coordServer with that name, if there
     * isn't you'd get a RemoteException. These examples hold true for
     * TestRegistry#getSimulator(String) and
     * TestRegistry#getVirtualServiceEnvironment(String). If you really want to fully
     * qualify the name you could pass "tcp://some.machine.com:2014/MySuperFastSimulator@SomeLab"
     *
     * @param vseName - the VSE to build a helper for
     *
     * @return a VseHelper
     */
    public static synchronized VseHelper getInstance( final String vseName )
    {
        if( !helpers.containsKey( vseName ) )
        {
            helpers.put( vseName, new VseHelper( vseName ) );
        }

        return helpers.get( vseName );
    }

    /** VirtualServiceEnvironment that this helper works with */
    private final VirtualServiceEnvironment vse;

    /**
     * Private constructor
     *
     * @param vseName - the VSE to find
     */
    private VseHelper( final String vseName )
    {
        vse = getVSE( vseName );
    }

    /**
     * Looks up the VirtualServiceEnvironment using the same rules
     * as TestRegistry#getVirtualServiceEnvironment(String).  See docs on getInstance(String)
     * for details.
     *
     * @param vseName - the vse to lookup
     *
     * @return VirtualServiceEnvironment
     */
    private static VirtualServiceEnvironment getVSE( final String vseName )
    {
        try
        {
            new Environment( "VSE Helper", Environment.APP_OTHER );

            String registryURL = NetConstants.getRegistryUrl();

            TestRegistry registry = Environment.setTestRegistry( registryURL );

            if( registry == null )
            {
                throw new Exception( "Registry '" + registryURL + "' not found" );
            }

            return registry.getVirtualServiceEnvironment( vseName );
        }
        catch( Exception e )
        {
            logger.log( Level.SEVERE, e.getMessage(), e );
            throw new RuntimeException( e.getMessage(), e );
        }
    }

    /** @return a Map of VirtualServiceStatus objects, keyed on service name */
    public Map<String, VirtualServiceStatus> getDeployedServiceInfo()
    {
        try
        {
            return vse.getServiceInfo();
        }
        catch( RemoteException e )
        {
            logger.log( Level.SEVERE, e.getMessage(), e );
            throw new RuntimeException( e.getMessage(), e );
        }
    }

    /** @return a list of deployed service names */
    public List<String> getDeployedServiceNames()
    {
        List<String> serviceNames = Lists.newArrayList();

        try
        {
            serviceNames.addAll( vse.getServiceNames() );
        }
        catch( RemoteException e )
        {
            logger.log( Level.SEVERE, e.getMessage(), e );
            throw new RuntimeException( e.getMessage(), e );
        }

        return serviceNames;
    }

    /** @return a unique list of model names */
    public List<String> getDeployedModelNames()
    {
        Set<String> modelNames = Sets.newHashSet();

        Map<String, VirtualServiceStatus> servicesMap = getDeployedServiceInfo();

        for( VirtualServiceStatus serviceInfo : servicesMap.values() )
        {
            modelNames.add( serviceInfo.getModelName() );
        }

        return new ArrayList<String>( modelNames );
    }

    /**
     * Get the MAR for the specified service from VSE
     *
     * @param serviceName - the Service
     *
     * @return the ModelArchive
     */
    public ModelArchive getMar( final String serviceName )
    {
        try
        {
            return vse.getModelArchive( serviceName );
        }
        catch( RemoteException e )
        {
            logger.log( Level.SEVERE, e.getMessage(), e );
            throw new RuntimeException( e.getMessage(), e );
        }
    }

    /** @param mar - the VS MAR to deploy */
    public void deployService( ModelArchive mar )
    {
        try
        {
            vse.addService( mar );
        }
        catch( Exception e )
        {
            logger.log( Level.SEVERE, e.getMessage(), e );
            throw new RuntimeException( e.getMessage(), e );
        }
    }

    /** @param mar - the VS MAR to deploy */
    public void redeployModel( ModelArchive mar )
    {
        try
        {
            vse.updateService( mar );
        }
        catch( Exception e )
        {
            logger.log( Level.SEVERE, e.getMessage(), e );
            throw new RuntimeException( e.getMessage(), e );
        }
    }

    /**
     * This method can be used to track what the status of the virtual service is.
     *
     * @param serviceName - the Service to check
     *
     * @return one of the following:
     *         <ul>
     *         <li>0 (VirtualService.STATUS_LOADED) the virtual service has been loaded.
     *         <li>1 (VirtualService.STATUS_STARTING) the virtual service has been asked to start but isn't
     *         there yet.
     *         <li>2 (VirtualService.STATUS_RUNNING) the virtual service is currently up and running.
     *         <li>3 (VirtualService.STATUS_STOPPING) the virtual service has been asked to stop but isn't
     *         there yet.
     *         <li>4 (VirtualService.STATUS_ENDED) the virtual service has ended.
     *         <li>5 (VirtualService.STATUS_NOT_RUNNABLE) the virtual service has errors that prevent it from being
     *         staged and executed.
     *         </ul>
     */
    public int getServiceStatus( String serviceName )
    {
        try
        {
            return vse.getService( serviceName ).getStatus();
        }
        catch( Exception e )
        {
            logger.log( Level.SEVERE, e.getMessage(), e );
            throw new RuntimeException( e.getMessage(), e );
        }
    }

    /** @param serviceName - the service to start */
    public void startService( String serviceName )
    {
        try
        {
            vse.startService( serviceName );
        }
        catch( Exception e )
        {
            logger.log( Level.SEVERE, e.getMessage(), e );
            throw new RuntimeException( e.getMessage(), e );
        }
    }

    /** @param serviceName - the service to stop */
    public void stopService( String serviceName )
    {
        try
        {
            vse.stopService( serviceName );
        }
        catch( Exception e )
        {
            logger.log( Level.SEVERE, e.getMessage(), e );
            throw new RuntimeException( e.getMessage(), e );
        }
    }

    /** @param serviceName - the service to remove from the environment */
    public void removeService( String serviceName )
    {
        try
        {
            vse.removeService( serviceName );
        }
        catch( Exception e )
        {
            logger.log( Level.SEVERE, e.getMessage(), e );
            throw new RuntimeException( e.getMessage(), e );
        }
    }

    /**
     * @param serviceName - the service for which the SI should be fetched
     *
     * @return a ServiceImage object (or null if no SI was found)
     */
    public ServiceImage getSiFromService( final String serviceName )
    {
        ServiceImage si;

        ModelArchive mar = getMar( serviceName );
        //si = ServiceImageHelper.getSIFromMar( mar );

        //return si;
        throw new UnsupportedOperationException();
    }

    /**
     * Merge the specified SI into the SI from the live service, then redeploy the live service
     *
     * @param siPath      - the source SI
     * @param serviceName - the target service
     * @param favorSource - whether or not to favor the source
     */
    public void mergeSiIntoLive( final String siPath, final String serviceName, boolean favorSource )
    {
        try
        {
            ModelArchive mar = getMar( serviceName );
            /*
            ServiceImage liveSI = ServiceImageHelper.getSIFromMar( mar );
            File targetSiFile = File.createTempFile( "vsi-", ".vsi" );
            ServiceImageHelper.saveServiceImage( liveSI, targetSiFile.getAbsolutePath() );

            ServiceImageHelper.combineServiceImages( siPath, targetSiFile.getAbsolutePath(), favorSource );
			
           // ModelArchive newMar = MarHelper.addAndreplaceFilesInMar( mar, new File[] { targetSiFile } );

            targetSiFile.delete();
            */

            //redeployModel( newMar );
        }
        catch( Exception e )
        {
            logger.log( Level.SEVERE, e.getMessage(), e );
            throw new RuntimeException( e.getMessage(), e );
        }
        throw new UnsupportedOperationException();
    }

    /**
     * Merge raw traffic into a live service
     *
     * @param rtPath - the raw traffic to use
     * @param serviceName - the service who's SI should be modified
     * @param vrsFile - the VRS file that provides the template for how to deal with stuff in the raw traffic file
     */
    public void mergeRawTrafficIntoLive( final String rtPath,
                                         final String serviceName,
                                         final String vrsFile )
    {
        try
        {
            ModelArchive mar = getMar( serviceName );
            /*
            ServiceImage liveSI = ServiceImageHelper.getSIFromMar( mar );
            ServiceImageHelper.mixRawTrafficIntoSi( rtPath, liveSI, vrsFile );

            File newSi = File.createTempFile( "vsi-", ".vsi" );
            ServiceImageHelper.saveServiceImage( liveSI, newSi.getAbsolutePath() );

            ModelArchive newMar = MarHelper.addAndreplaceFilesInMar( mar, new File[] { newSi } );

            newSi.delete();

            redeployModel( newMar );
            */
        }
        catch( Exception e )
        {
            logger.log( Level.SEVERE, e.getMessage(), e );
            throw new RuntimeException( e.getMessage(), e );
        }
        throw new UnsupportedOperationException();
    }

    /**
     * Merge raw traffic into a live service
     *
     * @param rtPath - the raw traffic to use
     * @param serviceName - the service who's SI should be modified
     * @param tph - the TransportProtocol to use when reading the Raw Traffic
     * @param reqDataProtocols - the request side DataProtocols to use when reading the raw traffic
     * @param rspDataProtocols - the response side DataProtocols to use when reading the raw traffic
     */
    public void mergeRawTrafficIntoLive( final String rtPath,
                                         final String serviceName,
                                         final TransportProtocol tph,
                                         final List<DataProtocol> reqDataProtocols,
                                         final List<DataProtocol> rspDataProtocols )
    {
        try
        {
            ModelArchive mar = getMar( serviceName );
            /*
            ServiceImage liveSI = ServiceImageHelper.getSIFromMar( mar );
            ServiceImageHelper.mixRawTrafficIntoSi( rtPath, liveSI, tph, reqDataProtocols, rspDataProtocols );

            File newSi = File.createTempFile( "vsi-", ".vsi" );
            ServiceImageHelper.saveServiceImage( liveSI, newSi.getAbsolutePath() );

            ModelArchive newMar = MarHelper.addAndreplaceFilesInMar( mar, new File[] { newSi } );

            newSi.delete();

            redeployModel( newMar );
            */
        }
        catch( Exception e )
        {
            logger.log( Level.SEVERE, e.getMessage(), e );
            throw new RuntimeException( e.getMessage(), e );
        }
        throw new UnsupportedOperationException();
    }

    /** @param vsmImportPath - the path from which VSMs should be read and deployed */
    public void deployAllServices( String vsmImportPath )
    {
        File vsmImportDir = new File( vsmImportPath );

        if( !vsmImportDir.isDirectory() || !vsmImportDir.exists() )
        {
            logger.severe( "vsmImportDir must point to a directory" );
        }

        for( File vsm : vsmImportDir.listFiles() )
        {
            if( vsm.isFile() )
            {
                String vsmPath = vsm.getAbsolutePath();

                try
                {
                    if( vsmPath.endsWith( ".vsm" ) )
                    {
                        logger.info( "Deploying VSM from " + vsmPath );
                        
                        //ModelArchive mar = TestCaseHelper.buildMarForVSM( vsmPath, null, 1, 100, true );
                        //String vsName = mar.getMARInfo().getName();

                        //deployService( mar );
                        //startService( vsName );
                    }
                }
                catch( Exception e )
                {
                    logger.log( Level.SEVERE, e.getMessage(), e );
                }
            }
        }
    }

    /** @param imageExportPath - the location where all SIs in VSE should be written to */
    public void exportAllServiceImages( String imageExportPath )
    {
        File basePathDir = new File( imageExportPath );
        if( !basePathDir.exists() )
        {
            if( !basePathDir.mkdirs() )
            {
                throw new RuntimeException( "Could not create output directory " + imageExportPath );
            }
        }

        List<String> serviceNames = getDeployedServiceNames();
        logger.log( Level.INFO, serviceNames.toString() );

        for( String serviceName : serviceNames )
        {
            logger.log( Level.INFO, "Exporting SI for service '" + serviceName + "' to " + imageExportPath );

            ServiceImage si = getSiFromService( serviceName );
            String siName = si.getName();
            String siPath = FileHelper.ensureTrailingFileSeparator( imageExportPath ) + siName + ".xml";
            
            //ServiceImageHelper.saveServiceImage( si, siPath );
        }
        throw new UnsupportedOperationException();
    }

    /** */
    public void removeAllServices()
    {
        List<String> serviceNames = getDeployedServiceNames();

        logger.log( Level.FINE, serviceNames.toString() );

        for( String serviceName : serviceNames )
        {
            logger.log( Level.INFO, "Removing " + serviceName );
            removeService( serviceName );
        }
    }

    /** */
    public void startAllServices()
    {
        List<String> serviceNames = getDeployedServiceNames();

        logger.log( Level.FINE, serviceNames.toString() );

        for( String serviceName : serviceNames )
        {
            logger.log( Level.INFO, "Starting " + serviceName );

            int status = getServiceStatus( serviceName );
            if( status != VirtualService.STATUS_RUNNING &&
                status != VirtualService.STATUS_STARTING &&
                status != VirtualService.STATUS_NOT_RUNNABLE )
            {
                startService( serviceName );
            }
        }
    }

    /** */
    public void stopAllServices()
    {
        List<String> serviceNames = getDeployedServiceNames();

        logger.log( Level.FINE, serviceNames.toString() );

        for( String serviceName : serviceNames )
        {
            logger.log( Level.INFO, "Stopping " + serviceName );

            int status = getServiceStatus( serviceName );
            if( status == VirtualService.STATUS_RUNNING ||
                status == VirtualService.STATUS_STARTING )
            {
                stopService( serviceName );
            }
        }
    }

    /**
     * This is a helper method for creating a transport protocol from its name.
     *
     * @param transportProtocol the name of the transport protocol to fetch
     * @param dataProtocols     an array of data protocols to use with the transport protocol
     * @param requestSide       a boolean for specifying whether we want request or response side data protocols
     *
     * @return the created protocol object.
     */
    public static TransportProtocol getTransportProtocol( String transportProtocol, String[] dataProtocols,
                                                          boolean requestSide )
    {
        ProtocolInfo info = transportProtocol == null ? null : getProtocolInfo( transportProtocol, false );

        if( info == null )
        {
            return null;
        }

        TransportProtocol protocol = info.createTransportProtocol();

        for( String dataProtocol : dataProtocols )
        {
            protocol.addDataProtocol( getProtocolInfo( dataProtocol, false ).createDataProtocol(), requestSide );
        }

        if( info.getType() == ProtocolInfo.Type.TRANSPORT && protocol.getDataProtocols( requestSide ).size() == 0 )
        {
            throw new IllegalArgumentException(
                                              "The \"" + transportProtocol +
                                              "\" transport protocol requires at least one data protocol." );
        }

        return protocol;
    }

    /**
     * @param transportProtocol the name of the transport protocol to fetch
     * @param dataProtocols     a Map with two keys ("request" and "response") who's values are data protocol maps
     *
     * @return the created protocol object
     */
    public static TransportProtocol getTransportProtocol( String transportProtocol,
                                                          Map<String, Map<String, ParameterList>> dataProtocols )
    {
        ProtocolInfo info = transportProtocol == null ? null : getProtocolInfo( transportProtocol, false );

        if( info == null )
        {
            return null;
        }

        TransportProtocol protocol = info.createTransportProtocol();

        Map<String, ParameterList> requestDPs = dataProtocols.get( "request" );
        for( String dph : requestDPs.keySet() )
        {
            DataProtocol dp = getProtocolInfo( dph, true ).createDataProtocol();
            if( requestDPs.get( dph ) != null )
            {
               // dp.initialize( requestDPs.get( dph ) );
            }
            protocol.addDataProtocol( dp, true );
        }

        Map<String, ParameterList> responseDPs = dataProtocols.get( "response" );
        for( String dph : responseDPs.keySet() )
        {
            DataProtocol dp = getProtocolInfo( dph, true ).createDataProtocol();
            if( responseDPs.get( dph ) != null )
            {
                //dp.initialize( responseDPs.get( dph ) );
            }
            protocol.addDataProtocol( dp, false );
        }

        if( info.getType() == ProtocolInfo.Type.TRANSPORT && protocol.getDataProtocols( true ).size() == 0 )
        {
            throw new IllegalArgumentException(
                                              "The \"" + transportProtocol +
                                              "\" transport protocol requires at least one data protocol." );
        }

        //return protocol;
        throw new UnsupportedOperationException();
    }

    /**
     * This is a helper method for converting a string into an appropriate protocol information object.  An exception
     * is
     * thrown if the name is not valid.
     *
     * @param name    the name of the desired protocol.
     * @param byClass whether the name is a class (true) or a human readable name (false)
     *
     * @return the protocol info object or {@code null}.
     */
    private static ProtocolInfo getProtocolInfo( String name, boolean byClass )
    {
        ProtocolInfo info;
        if( byClass )
        {
            info = ProtocolInfo.getProtocolInfo( name );
        }
        else
        {
            info = ProtocolInfo.getProtocolInfoByProtocolName( name );
        }

        if( info == null )
        {
            throw new IllegalArgumentException(
                                              "Unknown protocol: '" + name +
                                              "' - you need to quote names with spaces in them" );
        }

        return info;
    }
}
